Info << "Reading field T\n" << endl;

// pressure field
volScalarField p
(
    IOobject
    (
        "p",
        runTime.timeName(),
        mesh,
        IOobject::MUST_READ,
        IOobject::AUTO_WRITE
    ),
    mesh
);

// unity field used as a multiplier
volScalarField un
(
    IOobject
    (
        "un",
        runTime.timeName(),
        mesh,
        IOobject::MUST_READ,
        IOobject::NO_WRITE
    ),
    mesh
);

Info << "Reading transportProperties\n" << endl;

IOdictionary transportProperties
(
    IOobject
    (
        "transportProperties",
        runTime.constant(),
        mesh,
        IOobject::MUST_READ,
        IOobject::NO_WRITE
    )
);

// the nomenclature is provided in the transportProperties file located in the run directory
dimensionedTensor D1 (transportProperties.lookup("D1"));
dimensionedScalar T0  (transportProperties.lookup("T0"));
dimensionedScalar R  (transportProperties.lookup("R"));
dimensionedScalar rF_0 (transportProperties.lookup("rF_0"));
dimensionedScalar rF_min (transportProperties.lookup("rF_min"));
dimensionedScalar epsF_0 (transportProperties.lookup("epsF_0"));
dimensionedScalar OmegaF (transportProperties.lookup("OmegaF"));
dimensionedVector vg1 (transportProperties.lookup("vg1"));
dimensionedScalar kF (transportProperties.lookup("kF"));
dimensionedScalar p1 (transportProperties.lookup("p1"));
dimensionedScalar rampTime (transportProperties.lookup("rampTime"));

// Diffusion field (binary diffusion)
volTensorField D
(
    IOobject
    (
        "D",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    D1*un
);

// convection velocity
volVectorField vg
(
    IOobject
    (
        "vg",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    vg1*un
);

// norm of the convection velocity
volScalarField vgNorm
(
    IOobject
    (
        "vg",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    mag(vg1)*un
);

// mean fiber radius in the REV
volScalarField rF
(
    IOobject
    (
        "rF",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::AUTO_WRITE
    ),
    rF_0*un
);

// set initial fiber radius in the tube (ie, defining the location of the sample)
forAll(rF, i)
{
    if (mesh.C()[i][1] > (2.55 * 0.01)) {rF[i] = 0;}
    else {rF[i] = rF_0.value();}
}

Info << "rF" << nl << rF << nl;

scalar V_0 = 0;
forAll(rF, i)
{
    V_0 = V_0 + mesh.V()[i] * (rF[i] / rF_0.value());
}

Info << "V_0 = " << V_0 << nl;

// create the concentration field
volScalarField Xi
(
    IOobject
    (
        "Xi",
        runTime.timeName(),
        mesh,
        IOobject::MUST_READ,
        IOobject::AUTO_WRITE
    ),
    mesh
);
Xi = p1 * 0.2 / (R * T0) * 0;

/* Needed only for fixed gradient BC - not part of the official release because it is not robust (time step needs to be optimized)
    volScalarField gFN
    (
        IOobject
        (
            "gFN",
            runTime.timeName(),
            mesh,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        Xi*0
    );
*/

volVectorField gradXi
(
    IOobject
    (
        "gradXi",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::AUTO_WRITE
    ),
    fvc::grad(Xi)
);
// Read data
word movingPatchName(transportProperties.lookup("movingPatch"));
label movingPatchID = mesh.boundaryMesh().findPatchID(movingPatchName);

if (movingPatchID < 0)
{
    FatalErrorIn(args.executable())
            << "Problem with patch interpolation definition"
            << abort(FatalError);
}

// Read moving patch data
word movingRegionName(transportProperties.lookup("movingRegion"));

const fvMesh& motionMesh =
    runTime.objectRegistry::lookupObject<fvMesh>(movingRegionName);

/*** Imposing a fixedValue for Xi at the wall  ***/
// restricts to the moving part of the mesh
volScalarField& XiField =
    const_cast<volScalarField&>
    (
        motionMesh.objectRegistry::lookupObject<volScalarField>
        (
            "Xi"
        )
    );

// access to the right patch of the moving mesh
fixedValueFvPatchScalarField& XiMovingPatch =
    refCast<fixedValueFvPatchScalarField>
    (
        XiField.boundaryFieldRef()[movingPatchID]
    );

// update the pressure at the wall
scalarField XiPatchSize(XiMovingPatch.size());

