// Initialisation of chemistry operators for OpenFoam and Mutation++.
// The comments below give an overview of the initialisation and interfacing sequences.
// However, a fair knowledge of both code structures is obvisouly required to understand what follows in detail.

Info << "Initialization of chemistry operators and data\n" << endl;

// Read the chemistry property input directory for the ablative material (in OpenFoam)
IOdictionary chemistryProperties
(
    IOobject
    (
        "chemistryProperties",
        runTime.constant(),
        mesh,
        IOobject::MUST_READ,
        IOobject::NO_WRITE
    )
);

// Load the environementComposition file

fileName environmentCompositionDir(fileName(setCase.lookup("environmentComposition")).expand());

if (!isDir(environmentCompositionDir))
{
    Info << "Environment properties have not been found." << nl
	 << "Check file constant/setInput for appropriate path." << nl
         << "Exiting" << nl;
    return (0);
}

string copyEnvir("cp -rf " + environmentCompositionDir + "/. constant/environmentProperties");
system(copyEnvir);

IOdictionary environmentComposition
(
    IOobject
    (
        "constant/environmentProperties/environmentComposition",
        runTime,
        IOobject::MUST_READ,
        IOobject::NO_WRITE
    )
);

// Initialize finite-rate chemistry properties using OpenFoam finite-rate chemistry modules
Info << nl << "Reading thermophysicalProperties" << endl;
autoPtr<psiChemistryModel> pChemistry(psiChemistryModel::New(mesh));

psiChemistryModel& chemistry = pChemistry();
scalar dtChem = refCast<const psiChemistryModel>(chemistry).deltaTChem()[0];
scalar dtChemYEqn = 1e-6;
psiReactionThermo& thermo = chemistry.thermo();
thermo.validate(args.executable(), "h");

basicMultiComponentMixture& composition = thermo.composition();
PtrList<volScalarField>& Y = composition.Y(); // mass fraction of the species

word inertSpecie(thermo.lookup("inertSpecie"));

// Initialize Mutation++ to compute equilibrium chemistry, thermodynamics and transport properties
// Initialization of Mutation++ for equilibrium chemistry
word mixtureMutationEquilibrium (thermo.lookup("mixtureMutationEquilibrium"));
Mutation::MixtureOptions optEq(mixtureMutationEquilibrium);
optEq.setStateModel("Equil");
Mutation::Mixture mix(optEq);

const int ne_mix = mix.nElements();
// const int ns_mix = mix.nSpecies();
scalar p_Z[ne_mix];      // mass fractions of elements
scalar p_Z0[ne_mix];     // ref mass fractions of elements
scalar p_Zsub[ne_mix];   // mass fractions of elements in the subsurface cell
scalar p_Zx[ne_mix];     // mole fractions of elements
// scalar p_Xeq[ns_mix];    // mole fractions of species at equilibrium
// for(int i=0; i<ne_mix; i++) {p_Z[i]=mix.getDefaultComposition(i);} // composition given in mol fractions

Info << "Reading initial elemental composition of the gas in the sample" << nl;

for (int i = 0; i < ne_mix; i++)
{
    p_Zx[i] =
        constantProperties.lookupOrDefault<scalar>
        (
            "Zx[" + mix.elementName(i) + "]",
            0.0
        );
    Info << "Zx[" << mix.elementName(i) << "] = " << p_Zx[i] << nl;
    p_Z[i] = p_Zx[i];
}

mix.convert<Mutation::Thermodynamics::XE_TO_YE>(p_Z, p_Z);  // convert in mass fractions
for (int i = 0; i < ne_mix; i++)
{
    p_Z0[i] = p_Z[i];
    p_Zsub[i] = p_Z[i];
}

// Initialization of Mutation++ for surface equilibrium chemistry (B')
word mixtureMutationBprime(thermo.lookup("mixtureMutationBprime"));
Mutation::Mixture mixBprime(mixtureMutationBprime);
const int ne_Bprime = mixBprime.nElements();
const int ns_Bprime = mixBprime.nSpecies();

double* p_Yke = new double[ne_Bprime];
double* p_Ykg = new double[ne_Bprime];
double* p_Ykg_BC = new double[ne_Bprime];
double* p_Xw = new double[ns_Bprime];
double Bc_Mutation = 0.;
double hw_Mutation = 0.;
double Bg_Mutation = 0.;

// Default conditions (input in mole fractions -> converted in mass fractions)
Info << "Reading boundary-layer edge elemental composition" << nl;

for (int i = 0; i < ne_Bprime; i++)
{
    p_Yke[i] =
        environmentComposition.lookupOrDefault<scalar>
        (
            "Yke["+ mixBprime.elementName(i) + "]",
            0.0
        );
    Info << "Yke[" << mixBprime.elementName(i) << "] = " << p_Yke[i] << nl;
}

mixBprime.convert<Mutation::Thermodynamics::XE_TO_YE>(p_Yke, p_Yke);

Info << "Reading initial elemental composition of the pyrolysis gas for B' "
        "solver" << nl;

for (int i = 0; i < ne_Bprime; i++)
{
    p_Ykg[i] =
        constantProperties.lookupOrDefault<scalar>
        (
            "Zx[" + mix.elementName(i) + "]",
            0.0
        );
    Info << "Zx[" << mixBprime.elementName(i) << "] = " << p_Ykg[i] << nl;
}
// for(int i=0; i<ne_Bprime; i++) {p_Ykg[i]=mix.getDefaultComposition(i);}
mixBprime.convert<Mutation::Thermodynamics::XE_TO_YE>(p_Ykg, p_Ykg);

// Create and initialize the base fields (for each species/elements)
// Average multicomponent diffusion coefficients
PtrList<volScalarField> Dm(Y.size());

forAll(Dm, i)
{
    Dm.set
    (
        i,
        new volScalarField
        (
            IOobject
            (
                "Dm[" + std::to_string(i) + "](" + Y[i].name() + ")",
                runTime.timeName(),
                mesh,
                IOobject::NO_READ,
                IOobject::AUTO_WRITE
            ),
            mesh,
            D0
        )
    );
}
// Species enthalpy field
PtrList<volScalarField> hSpeciesField(Y.size());

forAll(hSpeciesField, i)
{
    hSpeciesField.set
    (
        i,
        new volScalarField
        (
            IOobject
            (
                "hSpeciesField[" + std::to_string(i) + "](" + Y[i].name() + ")",
                runTime.timeName(),
                mesh,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            hi
        )
    );
}

// Element mass-fraction fields
PtrList<volScalarField> Z(ne_mix);

forAll(Z, i)
{
    Z.set
    (
        i,
        new volScalarField
        (
            IOobject
            (
                "Z[" + mix.elementName(i) + "]",
                runTime.timeName(),
                mesh,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            Zdefault
        )
    );
}

// initialization of the element fractions - internal and fixed BC.
for (int i = 0; i < ne_mix; i++)
{
    Z[i] = p_Z0[i];

// jl - Sept 26, 2017  - The BC update is done in setElementsBCgrad.H in current version.
    forAll(Z[i].boundaryField()[movingPatchID], j)
    {
        Z[i].boundaryFieldRef()[movingPatchID][j] = p_Z0[i];
    }
// Therefore, this is overwritten. Non regression test is required before removing.
}

PtrList<surfaceVectorField> ZmassFlux(ne_mix);

forAll(ZmassFlux, i)
{
    ZmassFlux.set
    (
        i,
        new surfaceVectorField
        (
            IOobject
            (
                "ZmassFlux[" + mix.elementName(i) + "]",
                runTime.timeName(),
                mesh,
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            mDotGface
        )
    );
}

surfaceVectorField ZtotMassFlux// interpolation of mDotG on the faces
(
    IOobject
    (
        "ZtotMassFlux",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    mDotGface
);

multivariateSurfaceInterpolationScheme<scalar>::fieldTable fields;

forAll(Y, i)
{
    fields.add(Y[i]);
    fields.add(Dm[i]);
    fields.add(hSpeciesField[i]);
}

forAll(Z, i)
{
    fields.add(Z[i]);
}

forAll(Xsi, i)
{
    fields.add(Xsi[i]);
}

// effective pyrolyis gas flux (needed to compute convection in the species-conservation equations)
surfaceScalarField epsphi
(
    IOobject
    (
        "epsphi",
        runTime.timeName(),
        mesh,
        IOobject::READ_IF_PRESENT,
        IOobject::NO_WRITE
    ),
    mDotGface & mesh.Sf()
);


// Read pyrolysis-production SPECIES mass fractions - nb: only species present in the chemistry mechanism are looked up.
RectangularMatrix<scalar> Yp(sizeXsi, Y.size());

for (int i = 0; i < sizeXsi; i++)
{
    for (int j = 0; j < Y.size(); j++)
    {
        Yp[i][j] =
            constantProperties.lookupOrDefault<scalar>
            (
                "gamma[" + std::to_string(hashXsi[i][0]) + "][" +
                           std::to_string(hashXsi[i][1]) + "][" +
                           Y[j].name() + "]",
                0.0
            );
    }
}

// Read pyrolysis-production ELEMENT mass fractions - nb: only elements present
// in the equilibrium file (Mutation) are looked up.
RectangularMatrix<scalar> Zp(sizeXsi, ne_mix);

for (int i = 0; i < sizeXsi; i++)
{
    for (int j = 0; j < ne_mix; j++)
    {
        Zp[i][j] =
            constantProperties.lookupOrDefault<scalar>
            (
                "zeta[" + std::to_string(hashXsi[i][0]) + "][" +
                          std::to_string(hashXsi[i][1]) + "][" +
                          mix.elementName(j) + "]",
                0.0
            );
    }
}

// Initialisation of Mutation++ for finite-rate chemistry
word mixtureMutationFiniteRate(thermo.lookup("mixtureMutationFiniteRate"));
Mutation::MixtureOptions opts(mixtureMutationFiniteRate);
opts.setStateModel("ChemNonEq1T");
Mutation::Mixture mixFiniteRate(opts);

int speciesIndexMutation[Y.size()];
for (int i = 0; i < Y.size(); i++)
{
    speciesIndexMutation[i] = mixFiniteRate.speciesIndex(Y[i].name());

    if (finiteRate)
    {
        Info << "Y[" << i << "] = "
             << Y[i].name()
             << "(MutationIndex = "
             << mixFiniteRate.speciesIndex(Y[i].name())
             << ")"
             << nl;
    }
}

if (finiteRate)
{
    Info << "Species check .." << nl;

    for (int i = 0; i < Y.size(); i++)
    {
        if (Y[i].name() != mixFiniteRate.speciesName(speciesIndexMutation[i]))
        {
            Info << "Species name mismatch between PATO/constant/chemistryData "
                    "and Mutation/trunk/data/mixture.. this needs to be fixed "
                    "to run a finite-rate chemistry case."
                 << nl;

            Info << "PATO species " << i << " -> " << Y[i].name() << " = "
                 << mixFiniteRate.speciesName(speciesIndexMutation[i])
                 << "(MutationIndex = "
                 << mixFiniteRate.speciesIndex(Y[i].name()) << ")" << nl;
            return (0);
        }
    }
}

Info << ".. species correctly allocated" << nl;
const int ns  = mixFiniteRate.nSpecies();
double* pY = new double [ns];  // vector of species mass fraction in 0D - used in gasTT.H
double* pTp = new double [2]; // T, p vector in 0D - used in gasTT.H for setState(pTp, pY)

// extract indices of C(gr), O2 and CO2 for in-depth oxidation of the solid
// int iO2=0; int iCO2=0;
int iCs = 0;
for (int i = 0; i < Y.size(); i++) // if(Y[i].name() == "O2") iO2=i;
{
//    if(Y[i].name() == "CO2")
//    {
//        iCO2=i;
//    }

    if (Y[i].name() == "C(gr)")
    {
        iCs = i;
    }
}
dimensionedScalar M_Cs ("M_Cs", dimensionSet(1, 0, 0, 0, -1, 0, 0), 12.0112);

double* Ds = new double [ns]; // vector of species average diffusion coef in 0D - used in gasTT.H
double* hSpecies = new double [ns]; // vector of species enthalpies in 0D - used in YEqn.H

// Read boundary layer edge concentration (in MASS)

double* p_Yie = new double [ns];

Info << "Reading boundary-layer edge species composition" << nl;

for (int j = 0; j < Y.size(); j++)
{
    p_Yie[j] = 
        environmentComposition.lookupOrDefault<scalar>
        (
            "Yie[" + Y[j].name() + "]",
            0.0
        );
    
    Info << "Yie[" << Y[j].name() << "] = " <<  p_Yie[j] << nl;
}

// Allocations for sMass.H
dimensionedScalar rf0(constantProperties.lookup("rf0"));
dimensionedScalar rff(constantProperties.lookup("rff"));
dimensionedScalar rfFail(constantProperties.lookup("rfFail"));
int fiberPhase = constantProperties.lookupOrDefault<int>("fiberPhase", 0);
int matrixPhase = constantProperties.lookupOrDefault<int>("matrixPhase", 0);

volScalarField rT
(
    IOobject
    (
        "rT",
        runTime.timeName(),
        mesh,
        IOobject::MUST_READ,
        IOobject::AUTO_WRITE
    ),
    mesh
);

volVectorField grad_rT
(
    IOobject
    (
        "grad_rT",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    fvc::grad(rT)
);


volScalarField rT0
(
    IOobject
    (
        "rT0",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    rT
);

if (volAblation)
{
    // initial total radius based on the assumption that the matrix perfectly coats cylindrical fibers
    if (epsI[matrixPhase].value() != 0)
    {
        rT0 =
            std::sqrt
            (
                1. +
                epsI[fiberPhase].value() / epsI[matrixPhase].value()
            ) * rf0;
    }
    else
    {
        rT0 = rf0;
    }
    rT = rT0; // initialization
}

volScalarField specificSurface
(
    IOobject
    (
        "specificSurface",
        runTime.timeName(),
        mesh,
        IOobject::READ_IF_PRESENT,
        IOobject::NO_WRITE
    ),
    rT * 2. * eps[fiberPhase] / (rT0 * rT0)
);

volScalarField rho_ext
(
    IOobject
    (
        "rho_ext",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    rho[matrixPhase]
);

volScalarField siteDensity
(
    IOobject
    (
        "siteDensity",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    mesh,
    dimensionedScalar("zero", dimensionSet(0, -2, 0, 0, 1, 0, 0), 1.0)
);

volScalarField siteDensityIni
(
    IOobject
    (
        "siteDensityIni",
        runTime.timeName(),
        mesh,
        IOobject::NO_READ,
        IOobject::NO_WRITE
    ),
    mesh,
    dimensionedScalar("zero", dimensionSet(0, -2, 0, 0, 1, 0, 0), 1.0)
);

dimensionedScalar m_dot_ablation ("m_dot_ablation", dimensionSet(1, 0, -1, 0, 0, 0, 0), 0.0);

Info << "The chemistry properties have been read" << nl;
