// Update gas temperature (lagged in the energy conservation)
Tg = Ta; // if 1T model
T = Tg;  // T is used in the chemistry solver - better to use gas temperature

// Gas thermodynamics and transport properties
#include "updateDerivedFields/gasTT.H"

// Material conversion rate (from virgin to char)
// Material data are interpolated between virgin and char states
tau = 0. * tau;
forAll(Xsi, i)
{
    tau += normFp[i] * (1. - Xsi[i]);       // average advancement of the pyrolysis reactions
}
tau_w = tau.boundaryField()[movingPatchID]; // value at the wall (2D patch) - movingPatch is the ablated surface

// Geometrical properties
if(porous)
{
eps_g = eps_g_c + (eps_g_v - eps_g_c) * tau; // empirical porosity
K = K_c + (K_v - K_c) * tau;                 // permeability
}

// Densities
rho_g  = p * M / (R * Tg); // gas density (perfect gas law)
rho[0] = rho_g;
eps[0] = eps_g;

for (int i = 1; i < nSolidPhases + 1; i++) // update solid phase densities
{
    rho[i] = rhoI[i];
    forAll(Xsi, j)
    {
        if (hashXsi[j][0] == i)
            rho[i] -= rhoI[i] * Fp[j] * Xsi[j];
    }
}

rho_a = rho_a * 0.;
forAll(rho, i)
{
    rho_a += rho[i] * eps[i];    // effective density of the ablator
}

rho_s = rho_a - rho[0] * eps[0]; // effective density of the solid (excluding the gas)

// Material properties
// conductivity
forAll(kijk, i)
{
    kijk[i].xx() =
        kiCoef *
        (
            virgin.ki(p[i], Ta[i]) *
            tau[i] * rho_v / max(rho_s[i], rho_c)
          +
            charred.ki(p[i], Ta[i]) *
            (1. - tau[i] * rho_v / max(rho_s[i], rho_c))
        );
    kijk[i].yy() =
        kjCoef *
        (
            virgin.kj(p[i], Ta[i]) *
            tau[i] * rho_v / max(rho_s[i], rho_c)
          +
            charred.kj(p[i], Ta[i]) *
            (1. - tau[i] * rho_v / max(rho_s[i], rho_c))
        );
    kijk[i].zz() =
        kkCoef *
        (
            virgin.kk(p[i], Ta[i]) *
            tau[i] * rho_v / max(rho_s[i], rho_c)
          +
            charred.kk(p[i], Ta[i]) *
            (1. - tau[i] * rho_v / max(rho_s[i], rho_c))
        );
}

forAll(kijk.boundaryField()[movingPatchID], faceI) // faceI refers to the faces in the patch (ie, loop on all the faces of the patch for 3D problems - only 1 face in 1D)
{
    kijk.boundaryFieldRef()[movingPatchID][faceI].xx() =
        kiCoef *
        (
            virgin.ki(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) *
            tau_w[faceI] * rho_v / max(rho_s.boundaryField()[movingPatchID][faceI], rho_c)
          +
            charred.ki(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) *
            (1. - tau_w[faceI] * rho_v / max(rho_s.boundaryField()[movingPatchID][faceI], rho_c))
        );
    kijk.boundaryFieldRef()[movingPatchID][faceI].yy() =
        kjCoef *
        (
            virgin.kj(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) *
            tau_w[faceI] * rho_v / max(rho_s.boundaryField()[movingPatchID][faceI], rho_c)
          +
            charred.kj(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) *
            (1. - tau_w[faceI] * rho_v / max(rho_s.boundaryField()[movingPatchID][faceI], rho_c))
        );
    kijk.boundaryFieldRef()[movingPatchID][faceI].zz() =
        kkCoef *
        (
            virgin.kk(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) *
            tau_w[faceI] * rho_v / max(rho_s.boundaryField()[movingPatchID][faceI], rho_c)
          +
            charred.kk(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) *
            (1. - tau_w[faceI] * rho_v / max(rho_s.boundaryField()[movingPatchID][faceI], rho_c))
        );
}
k = tP & kijk & tP.T(); // passage from basis ijk (main directions of the conductivity tensor) to basis xyz (mesh)
k_abl_sym = symm(k);

// heat capacity of the solid
forAll (rhoCp, i)
{
    rhoCp[i] =
        rho_s[i] *
        (
            virgin.cp(p[i], Ta[i]) *
            tau[i] * rho_v / max(rho_s[i], rho_c)
          +
            charred.cp(p[i], Ta[i]) *
            (1. - tau[i] * rho_v / max(rho_s[i], rho_c))
        );
}

forAll(rhoCp.boundaryField()[movingPatchID], faceI)
{
    rhoCp.boundaryFieldRef()[movingPatchID][faceI] =
        rho_s.boundaryField()[movingPatchID][faceI] *
        (
            virgin.cp(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) *
            tau_w[faceI] * rho_v / max(rho_s.boundaryField()[movingPatchID][faceI], rho_c)
          +
            charred.cp(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) *
            (1. - tau_w[faceI] * rho_v / max(rho_s.boundaryField()[movingPatchID][faceI], rho_c))
        );
}

// heat of pyrolysis
if (rho_c < rho_v)
{
    forAll (h_bar, i)
    {
        h_bar[i] =
            (rho_v * virgin.h(p[i], Ta[i]) - rho_c * charred.h(p[i], Ta[i])) /
            (rho_v - rho_c);
    }

    forAll (h_bar.boundaryField()[movingPatchID], faceI)
    {
        h_bar.boundaryFieldRef()[movingPatchID][faceI] =
            (
                rho_v * virgin.h(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]) -
                rho_c * charred.h(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI])
            ) /
            (rho_v - rho_c);
    }

    if (!detailedSolidEnthalpies)
    {
        pyrolysisEnergy = - h_bar * piTotal;    // NASA Apollo approach (Kendall 1960)
    }
    else
    {
        pyrolysisEnergy = pyrolysisEnergy * 0.; // Multiphase approach (Lachaud 2017)
        volScalarField hs = h_bar * 0.;
        forAll (h_bar, i)
        {
            hs[i] =
                (
                    rho_v * (virgin.hs(p[i], Ta[i]) - virgin.hs(p[i], 298)) -
                    rho_c * (charred.hs(p[i], Ta[i]) - charred.hs(p[i], 298))
                ) /
                (rho_v - rho_c);
        }

        forAll(pi, i)
        {
            pyrolysisEnergy += - pi[i] * (hp[i] + hs);
        }
    }
}

// char enthalpy
forAll (h_c, i)
{
    h_c[i] = charred.h(p[i], Ta[i]);
}

forAll (h_c.boundaryField()[movingPatchID], faceI)
{
    h_c.boundaryFieldRef()[movingPatchID][faceI] =
        charred.h(p.boundaryField()[movingPatchID][faceI], Ta.boundaryField()[movingPatchID][faceI]);
}

// enthalpy of the solid lost with heterogeneous reactions
omegaHeterogeneousEnergy = - h_c * omegaHeterogeneousTotal;

// velocity
if (writeVelocityComponents)
{
    vi = v.component(0);
    vj = v.component(1);
    vk = v.component(2);
    vmag = mag(v);
}
// mean gas velocity
if (cfd)
{
    U = v * eps_g;
/*    
forAll(mesh.boundaryMesh(), patchouli)
{
// store surface normal vectorField of patch faces
    vectorField nf =
     mesh.Sf().boundaryField()[patchouli]
    / mesh.magSf().boundaryField()[patchouli];
// project velocity normal on patch faces
    U.boundaryFieldRef()[patchouli] =  (U.boundaryField()[patchouli] & nf) * nf;
}
*/
}

